"""
CJKSplitter - Chinese, Japanese, Korean word splitter for ZCTextIndex

(C) by www.ZopeChina.com, panjy@plone.cn & others

License: see LICENSE.txt

$Id: Exp $

DESCRIPTION OF ALGORITHM

  Continuous letters and numbers make up words.  Spaces and symbols
  separate letters and numbers into words.  This is sufficient for
  all western text.

  CJK doesn't use spaces or separators to separate words, so the only
  way to really find out what constitutes a word would be to have a
  dictionary and advanced heuristics.  Instead, we form pairs from
  consecutive characters, in such a way that searches will find only
  characters that appear more-or-less the right sequence.  For example:

    ABCDE => AB BC CD DE

  This works okay since both the index and the search query is split
  in the same manner, and since the set of characters is huge so the
  extra matches are not significant.
"""

from Products.ZCTextIndex.ISplitter import ISplitter
from Products.ZCTextIndex.PipelineFactory import element_factory
import re

def getSupportedEncoding(encodings):
    for encoding in encodings:
        try:
            unicode('A', encoding)
            return encoding
        except:
            pass
    return 'utf-8'

class CJKSplitter:
    rx = re.compile(u"[a-zA-Z0-9_]+|[\u4E00-\u9FFF]+", re.UNICODE)
    rxGlob = re.compile(u"[a-zA-Z0-9_]+[*?]*|[\u4E00-\u9FFF]+", re.UNICODE)

    default_encoding = "utf-8"

    def process(self, lst, isGlob=0):
        result = []
        if isGlob:
          rx = self.rxGlob
        else:
          rx = self.rx
        for s in lst:
            if type(s) is type(''): # not unicode
                s = unicode(s, self.default_encoding, 'replace')
            splitted = rx.findall(s)
            for w in splitted:
                if ord(w[0]) >= ord(u'\u4E00'):  # CJK words?
                    len_w = len(w)
                    if len_w == 2:
                        result.append(w)
                    else:
                        for i in range(1, len_w - 1):
                            result += [w[i-1:i+1], w[i:i+2]]
                else:
                    result.append(w)
        return result

    def processGlob(self, lst):
        return self.process(lst, 1)

gb_encoding = getSupportedEncoding(['gb18030', 'mbcs', 'gbk', 'gb2312']) 
class GBSplitter(CJKSplitter):
    default_encoding = gb_encoding

big5_encoding = getSupportedEncoding(['big5', 'mbcs'])
class BIG5Splitter(CJKSplitter):
    default_encoding = big5_encoding

try:
    element_factory.registerFactory('Word Splitter',
          'CJK splitter', CJKSplitter)
    element_factory.registerFactory('Word Splitter',
          'CJK GB splitter', GBSplitter)
    element_factory.registerFactory('Word Splitter',
          'CJK BIG5 splitter', BIG5Splitter)
except ValueError:
    # in case the splitter is already registered, ValueError is raised
    pass

